"""remarx: Marx quote identification for the CDH Citing Marx project."""

# Copyright (c) 2025, Center for Digital Humanities, Princeton University
# SPDX-License-Identifier: Apache-2.0

from remarx import app, sentence

__version__ = "0.3rc1"

__all__ = ["__version__", "app", "sentence"]
